#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
A command that generates various statistics.
'''

# TODO This has to be migrated to v1. This was not done immediatly, because we wanted
# to revise the statistics anyways: https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-FAIR/-/issues/611

# import matplotlib.scale as mscale
# import matplotlib.transforms as mtransforms
# import matplotlib.pyplot as plt
# import matplotlib.ticker as ticker
# import numpy as np
# import click
# import json
# import datetime
# import subprocess

# from nomad import config
# from .client import client


# def codes(client, minimum=1, **kwargs):
#     data = client.repo.search(per_page=1, **kwargs).response().result

#     x_values = sorted([
#         code for code, values in data.statistics['dft.code_name'].items()
#         if code != 'not processed' and values.get('calculations', 1000) >= minimum], key=lambda x: x.lower())

#     return data.statistics, x_values, 'code_name', 'code'


# def dates(client, minimum=1, **kwargs):
#     data = client.repo.search(per_page=1, date_histogram=True, **kwargs).response().result

#     x_values = list([
#         x for x in data.statistics['date_histogram'].keys()])

#     return data.statistics, x_values, 'date_histogram', 'month'


# def error_fig(client):
#     _, labels, _, _ = codes(client)

#     def code_values(metric='code_runs', **kwargs):
#         result = client.repo.search(
#             per_page=1,
#             owner='admin',
#             metrics=[] if metric == 'code_runs' else metric,
#             **kwargs).response().result

#         return {
#             code: values[metric]
#             for code, values in result.quantities['dft.code_name'].items()
#             if code != 'not processed' and (not labels or code in labels) > 0}

#     # get the data
#     all_entries = code_values()
#     parser_failure_label = 'parser failure'
#     error_types = [
#         {'name': parser_failure_label, 'search': dict(system='not processed')},
#         {'name': 'failed system classification', 'search': dict(system='unavailable')},
#         {'name': 'no basis set available', 'search': dict(basis_set='unavailable')},
#         {'name': 'no XC functional available', 'search': dict(xc_functional='unavailable')}
#     ]
#     errors = {
#         error_type['name']: {
#             code: failures
#             for code, failures in code_values(**error_type['search']).items()}
#         for error_type in error_types}
#     errors_rates = {
#         error_type['name']: {
#             code: 0 if all_entries[code] == 0 else failures / all_entries[code]
#             for code, failures in code_values(**error_type['search']).items()}
#         for error_type in error_types}

#     fig, axs = plt.subplots(figsize=(15, 12), dpi=72, nrows=2)

#     def draw_error_chart(errors, ax, colors, entries=None, mul=1, scale=0.5):
#         n_bars = len(errors) - 1
#         leg_colors = list(colors)

#         x = np.arange(len(labels))  # the label locations
#         width = 0.7 / n_bars  # the width of the bars
#         plt.sca(ax)
#         plt.xticks(rotation=90)

#         if entries is not None:
#             ax.bar(x, [entries[code] for code in labels], width * n_bars, label='all entries', color=colors.pop(0))

#         i = -1
#         not_processed = [errors[parser_failure_label][code] * mul for code in labels]
#         ax.bar(x, not_processed, width * n_bars, label=parser_failure_label, color=colors.pop(0))
#         for key, values in errors.items():
#             if key != parser_failure_label:
#                 ax.bar(x + i * width, [values[code] * mul for code in labels], width, label=key, bottom=not_processed, color=colors.pop(0))
#                 i += 1

#         # Add some text for labels, title and custom x-axis tick labels, etc.
#         ax.set_yscale('power', exponent=scale)
#         ax.set_xticks(x)
#         ax.set_xticklabels(labels)
#         ax.legend()
#         leg = ax.get_legend()
#         for i in range(0, len(leg_colors)):
#             leg.legendHandles[i].set_color(leg_colors[i])

#         fig.tight_layout()

#     ax = axs[0]
#     ax.set_title('Absolute number of entries with parser errors or missing repository metadata compared to all entries per code')
#     ax.set_ylabel('number of entries', )
#     colors = ['grey', 'red', 'yellow', 'orange', 'brown']
#     draw_error_chart(errors, ax, entries=all_entries, mul=1, scale=0.25, colors=colors)
#     ax.yaxis.set_major_formatter(ticker.StrMethodFormatter('{x:,.0f}'))

#     ax = axs[1]
#     ax.set_title('Relative rates of entries with parser errors or missing repository metadata per code')
#     ax.set_ylabel('rate in %', )
#     colors = ['red', 'yellow', 'orange', 'brown']
#     draw_error_chart(errors_rates, ax, mul=100, colors=colors)

#     plt.show()

#     return fig, plt


# class Metric:
#     def __init__(self, metric, label=None, power=None, multiplier=1, format=None, cumulate=False):
#         if label is None:
#             label = metric

#         self.metric = metric
#         self.agg = None
#         self.label = label
#         self.multiplier = multiplier
#         self.power = power
#         self.format = format
#         self.cumulate = cumulate

#     def draw_axis(self, axis, data, x_values, x_positions, width, color, only=False):
#         label_color = 'black' if only else color
#         value_map = {
#             x: values[self.metric]
#             for x, values in data[self.agg].items()
#             if x in x_values}

#         if self.power is not None:
#             axis.set_yscale('power', exponent=self.power)
#         else:
#             axis.set_yscale('log')
#         axis.set_ylabel(self.label, color=label_color)

#         if self.format is not None:
#             axis.yaxis.set_major_formatter(ticker.StrMethodFormatter(self.format))

#         y_values = [value_map[x] * self.multiplier for x in x_values]
#         if self.cumulate:
#             y_values = np.array(y_values).cumsum()
#         axis.bar(x_positions, y_values, width, label=self.label, color=color, align='edge')
#         axis.tick_params(axis='y', labelcolor=label_color)

#         # TODO remove
#         # for x, v in zip(x_positions, y_values):
#         #     axis.text(x + .1, v, ' {:,}'.format(int(v)), color=color, fontweight='bold', rotation=90)
#         # import matplotlib.lines as mlines
#         # line = mlines.Line2D([min(x_positions), max(x_positions)], [72, 72], color=color)
#         # axis.add_line(line)


# def bar_plot(
#         client, retrieve, metric1, metric2=None, title=None, format_xlabel=None,
#         xlim={}, ylim=dict(bottom=1), **kwargs):
#     if format_xlabel is None:
#         format_xlabel = lambda x: x

#     metrics = [] if metric1.metric == 'code_runs' else [metric1.metric]
#     if metric2 is not None:
#         metrics += [] if metric2.metric == 'code_runs' else [metric2.metric]

#     data, x_values, agg, agg_label = retrieve(client, metrics=metrics, statistics=['dft.code_name'], **kwargs)
#     metric1.agg = agg
#     if metric2 is not None:
#         metric2.agg = agg

#     fig, ax1 = plt.subplots(figsize=(7, 4), dpi=72)
#     x = np.arange(len(x_values))
#     width = 0.8 / 2
#     if metric2 is None:
#         width = 0.8
#     plt.sca(ax1)
#     plt.xticks(rotation=90)
#     ax1.set_xticks(x)
#     ax1.set_xticklabels([format_xlabel(value) if value != 'Quantum Espresso' else 'Q. Espresso' for value in x_values])
#     ax1.margins(x=0.01)
#     ax1.set_xlim(**xlim)
#     # i = 0
#     # for label in ax1.xaxis.get_ticklabels():
#     #     label.set_visible(i % 4 == 0)
#     #     i += 1

#     if title is None:
#         title = 'Number of %s' % metric1.label
#         if metric2 is not None:
#             title += ' and %s' % metric2.label
#         title += ' per %s' % agg_label
#         ax1.set_title(title)
#     elif title != '':
#         ax1.set_title(title)

#     metric1.draw_axis(ax1, data, x_values, x - (width / 2), width, 'tab:blue', only=metric2 is None)
#     ax1.set_ylim(**ylim)
#     # ax1.set_yticks([40, 30, 20, 10, 5, 1, 0.5, 0.1])
#     if not metric2:
#         ax1.grid(which='major', axis='y', linestyle='-')

#     if metric2:
#         ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
#         metric2.draw_axis(
#             ax2, data, x_values, x + width / 2, width, 'tab:red')
#         ax2.set_ylim(bottom=1)

#     fig.tight_layout()

#     return fig, plt


# @client.command(help='Generate various matplotlib charts')
# @click.option('--errors', is_flag=True, help='Two charts with relative and absolute parser/normalizer errors per code.')
# @click.option('--x-axis', type=str, help='Aggregation used for x-axis, values are "code" and "time".')
# @click.option('--y-axis', multiple=True, type=str, help='Metrics used for y-axis, values are "entries", "energies", "calculations", "users".')
# @click.option('--cumulate', is_flag=True, help='Cumulate over x-axis.')
# @click.option('--title', type=str, help='Override chart title with given value.')
# @click.option('--total', is_flag=True, help='Provide total sums of key metrics.')
# @click.option('--save', type=str, help='Save to given file instead of showing the plot.')
# @click.option('--power', type=float, help='User power scale instead of log with the given inverse power.')
# @click.option('--open-access', is_flag=True, help='Only consider Open-Access data.')
# @click.option('--minimum', type=int, default=1, help='Only consider codes with at least the given ammount of entries.')
# def statistics_plot(errors, title, x_axis, y_axis, cumulate, total, save, power, open_access, minimum):
#     from .client import create_client
#     client = create_client()

#     class PowerScale(mscale.ScaleBase):
#         name = 'power'

#         def __init__(self, axis, exponent, **kwargs):
#             mscale.ScaleBase.__init__(self, axis, **kwargs)
#             self.exponent = exponent

#         def set_default_locators_and_formatters(self, axis):
#             axis.set_major_locator(ticker.AutoLocator())
#             axis.set_major_formatter(ticker.ScalarFormatter())
#             axis.set_minor_locator(ticker.NullLocator())
#             axis.set_minor_formatter(ticker.NullFormatter())

#         def limit_range_for_scale(self, vmin, vmax, minpos):
#             return max(0., vmin), vmax

#         class Transform(mtransforms.Transform):
#             input_dims = 1
#             output_dims = 1
#             is_separable = True

#             def __init__(self, exponent):
#                 super().__init__()
#                 self.exponent = exponent

#             def transform_non_affine(self, a):
#                 return np.array(a)**self.exponent

#             def inverted(self):
#                 return PowerScale.Transform(1 / self.exponent)

#         def get_transform(self):
#             return self.Transform(self.exponent)

#     mscale.register_scale(PowerScale)

#     kwargs = {}
#     if cumulate:
#         kwargs.update(
#             power=1,
#             multiplier=1e-6,
#             format='{x:,.1f}M')
#     elif power is not None:
#         kwargs.update(
#             power=1 / power,
#             multiplier=1e-6,
#             format='{x:,.1f}M')

#     metrics = {
#         'entries': Metric(
#             'code_runs',
#             label='entries (code runs)',
#             cumulate=cumulate,
#             **kwargs),
#         'users': Metric(
#             'users',
#             cumulate=cumulate,
#             label='users that provided data'),
#         'energies': Metric(
#             'total_energies',
#             label='total energy calculations',
#             cumulate=cumulate,
#             **kwargs),
#         'calculations': Metric(
#             'calculations',
#             label='calculations (e.g. total energy)',
#             cumulate=cumulate,
#             **kwargs)
#     }

#     if errors:
#         fig, plt = error_fig(client)

#     owner = 'all' if open_access else 'admin'

#     if x_axis is not None:
#         assert 1 <= len(y_axis) <= 2, 'Need 1 or 2 y axis'

#         kwargs = {}
#         if x_axis == 'code':
#             x_axis = codes
#             kwargs.update(ylim=dict(bottom=1))
#         elif x_axis == 'time':
#             x_axis = dates
#             kwargs.update(
#                 ylim=dict(bottom=1),
#                 format_xlabel=lambda x: datetime.datetime.fromtimestamp(int(x) / 1000).strftime('%b %y'))
#         else:
#             assert False, 'x axis can only be "code" or "time"'

#         y_axis = [metrics[y] for y in y_axis]

#         fig, plt = bar_plot(
#             client, x_axis, *y_axis, title=title, owner=owner, minimum=minimum, **kwargs)

#     if errors or x_axis is not None:
#         if save is not None:
#             fig.savefig(save, bbox_inches='tight')
#         else:
#             plt.show()

#     if total:
#         data = client.repo.search(
#             per_page=1, owner=owner, statistics=True,
#             metrics=['total_energies', 'calculations', 'uploaders', 'authors', 'datasets']).response().result
#         print(json.dumps(data.statistics['total'], indent=4))


# @client.command(help='Generate table with basic statistics summary.')
# @click.option('--html', is_flag=True, help='Output HTML instead of plain text table.')
# @click.option('--geometries', is_flag=True, help='Use geometries not unique geometries.')
# @click.option('--public-path', type=str, default=config.fs.public, help='The path to the public data. Default is %s.' % config.fs.public)
# def statistics_table(html, geometries, public_path):
#     # get more stats for files
#     # uploads: find . -maxdepth 2 | wc -l
#     # public archive: find . -regex '.*archive.*public.*zip' -type f -print0 | du --files0-from=- -ch | grep total$
#     # public raw: find . -regex '.*raw.*public.*zip' -type f -print0 | du --files0-from=- -ch | grep total$

#     def get_statistic(response, quantity, value, metric):
#         quantity_data = response.statistics.get(quantity)
#         if quantity_data is None:
#             return 0
#         value_data = quantity_data.get(value)
#         if value_data is None:
#             return 0

#         value = value_data.get(metric)
#         return value if value is not None else 0

#     from nomad.cli.client import create_client
#     client = create_client()

#     geometry_metric = 'dft.unique_geometries' if not geometries else 'dft.geometries'

#     # search scc with system type
#     data_all = client.repo.search(
#         per_page=1, metrics=['dft.calculations'], statistics=['dft.system', 'dft.code_name']).response().result

#     entries = get_statistic(data_all, 'total', 'all', 'code_runs')
#     calculations = get_statistic(data_all, 'total', 'all', 'dft.calculations')
#     calculations_1d = get_statistic(data_all, 'dft.system', '1D', 'dft.calculations') \
#         + get_statistic(data_all, 'dft.system', 'atom', 'dft.calculations') \
#         + get_statistic(data_all, 'dft.system', 'molecule / cluster', 'dft.calculations')

#     calculations_2d = get_statistic(data_all, 'dft.system', '2D / surface', 'dft.calculations')
#     calculations_2d += get_statistic(data_all, 'dft.system', '2D', 'dft.calculations')
#     calculations_2d += get_statistic(data_all, 'dft.system', 'surface', 'dft.calculations')
#     calculations_3d = get_statistic(data_all, 'dft.system', 'bulk', 'dft.calculations')

#     metrics_all = client.repo.search(per_page=1, metrics=[geometry_metric, 'dft.quantities']).response().result
#     geometries = get_statistic(metrics_all, 'total', 'all', geometry_metric)
#     quantities = get_statistic(metrics_all, 'total', 'all', 'dft.quantities')

#     # search calcs quantities=section_k_band
#     band_structures = get_statistic(
#         client.repo.search(per_page=1, **{'dft.quantities': ['section_k_band']}).response().result,
#         'total', 'all', 'code_runs')

#     # search calcs quantities=section_dos
#     dos = get_statistic(
#         client.repo.search(per_page=1, **{'dft.quantities': ['section_dos']}).response().result,
#         'total', 'all', 'code_runs')

#     phonons = get_statistic(
#         client.repo.search(per_page=1, **{'dft.code_name': 'Phonopy'}).response().result,
#         'total', 'all', 'code_runs')

#     # files and sized
#     def run_shell_command(command):
#         process = subprocess.run(['bash', '-c', command], stdout=subprocess.PIPE)
#         out = process.stdout.decode('utf-8')
#         if process.stderr is not None:
#             err = process.stderr.decode('utf-8')
#             print('There is an error: %s' % str(err.strip()))
#         return out.split('\t')[0].strip()

#     archive_data = run_shell_command((
#         'find %s -regex \'.*archive.*public.*zip\' '
#         '-type f -print0 | du --files0-from=- -ch | grep total$') % public_path)
#     raw_data = run_shell_command((
#         'find %s -regex \'.*raw.*public.*zip\' '
#         '-type f -print0 | du --files0-from=- -ch | grep total$') % public_path)
#     n_uploads = run_shell_command(
#         'find %s -regex \'.*raw.*public.*zip\' -type f | wc -l' % public_path)

#     try:
#         n_uploads = '{:,}'.format(int(n_uploads))
#     except Exception:
#         pass

#     if not html:
#         print('''
#             Entries: {:,.0f}
#             Calculations, e.g. total energies: {:,.0f}
#             Geometries: {:,.0f}
#             Bulk crystals: {:,.0f}
#             2D / Surfaces: {:,.0f}
#             Atoms / Molecules: {:,.0f}
#             DOS: {:,.0f}
#             Band structures: {:,.0f}
#             Total parsed quantities: {:,.0f}
#             Public raw data: {}B
#             Public archive data: {}B
#             Number of uploads: {}
#         '''.format(
#             entries,
#             calculations,
#             geometries,
#             calculations_3d,
#             calculations_2d,
#             calculations_1d,
#             dos,
#             band_structures,
#             quantities,
#             raw_data,
#             archive_data,
#             n_uploads
#         ))

#     else:
#         print('''
#             <div class="container">
#                 <p>The <i>NOMAD Archive</i> stores calculations performed
#                 with all the most important and widely used electronic-structure and force-field codes
#                 in a code-independent format.
#                 </p>
#                 <p>Summary statistics of the Archive content (last update in {}):</p>
#                 <table class="table" style="text-align: left; max-width: 700px;">
#                     <thead>
#                         <tr>
#                         <th scope="col">Metric</th>
#                         <th scope="col">Value</th>
#                         </tr>
#                     </thead>
#                     <tbody>
#                         <tr>
#                         <td scope="row">Entries, i.e. code runs</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Calculations, e.g. total energies</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Geometries</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Bulk Crystals</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Surfaces</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Molecules/Clusters</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">DOS</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Band Structures</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Phonon Calculations</th>
#                         <td>{:,.0f}</td>
#                         </tr>
#                         <tr>
#                         <td scope="row">Overall parsed quantities</td>
#                         <td>{:,.0f}</td>
#                         </tr>
#                     </tbody>
#                 </table>
#                 <p>
#                     Furthermore:
#                 </p>
#                 <ul>
#                     <li><b>{}</b> Uploads with <b>{}B</b> of raw data</li>
#                     <li><b>{}B</b> of archive data</li>
#                     <li>Data classified using <b>168</b> public metadata of the NOMAD Meta Info and <b>2,360</b> code-specific metadata</li>
#                 </ul>
#                 <p>
#                     For more and interactive statistics, use the <i>metadata</i> view of
#                     the <a href="https://nomad-lab.eu/prod/rae/gui/search">NOMAD Repository and Archvi search</a>.
#                 </p>
#                 <p>
#                     90% of VASP calculations are provided by
#                         <a href="http://aflowlib.org">AFLOWlib</a> (S. Curtarolo),
#                         <a href="http://oqmd.org"> OQMD</a> (C. Wolverton) and
#                         <a href="https://materialsproject.org">Materials Project</a> (K. Persson).
#                 </p>
#                 <p>
#                     You can further explore the statistics in the below dynamic histograms. To
#                     change the displayed quantity, select from the "Quantities" drop-down. To
#                     filter the data, click histogram bars for different filter combinations.
#                     To reset filters, click "Reset Filters".
#                 </p>
#                 <p>
#                     The archive data is represented in a code-independent, structured
#                     form. The archive structure and all quantities are described via the
#                     <a href="https://nomad-lab.eu/prod/rae/gui/metainfo">NOMAD Metainfo</a>.
#                     The NOMAD Metainfo defines a conceptual model to store the values connected
#                     to atomistic or <i>ab initio</i> calculations. A clear and usable metadata definition
#                     is a prerequisites to preparing the data for analysis that everybody
#                     can contribute to.
#                 </p>
#                 <p>
#                     In collaboration with the <a href="http://www.bbdc.berlin/">Berlin Big Data Center (BBDC)</a>,
#                     we use the Apache Flink infrastructure to support and go beyond the standard MapReduce model to enable
#                     rapid and complex queries.
#                 </p>
#                 <p>
#                     Contact concerning general aspects of the CoE: <a href="mailto:pietsch@fhi-berlin.mpg.de">Jessica Pietsch</a>
#                 </p>
#                 <p>
#                     Contact concerning the NOMAD Archive:
#                         <a href="mailto:markus.scheidgen@physik.hu-berlin.de">Markus Scheidgen</a>,
#                         <a href="mailto:ghiringhelli@fhi-berlin.mpg.de">Luca Ghiringhelli</a>
#                 </p>
#             </div>
#         '''.format(
#             datetime.datetime.now().strftime('%b %Y'),
#             entries,
#             calculations,
#             geometries,
#             calculations_3d,
#             calculations_2d,
#             calculations_1d,
#             dos,
#             band_structures,
#             phonons,
#             quantities,
#             n_uploads,
#             raw_data,
#             archive_data
#         ))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module comprises the nomad@FAIRDI APIs. Currently there is NOMAD's official api, optimade api,
and dcat api. The app module also servers documentation, gui, and
alive.
'''
from flask import Flask, jsonify, url_for, request
from flask_cors import CORS

from nomad import config
from nomad.processing import Upload
from nomad.infrastructure import keycloak

from .docs import blueprint as docs_blueprint
from .dist import blueprint as dist_blueprint
from .gui import blueprint as gui_blueprint

from h5grove.flaskutils import BLUEPRINT as h5grove_blueprint


def auth_before_request():
    """Checks whether the user is authenticated and has read access to the upload."""
    token = request.args.get("token")
    user = keycloak.tokenauth(token)

    upload_id = request.args.get("upload_id")
    upload = Upload.objects(upload_id=upload_id).first()
    if not (user and (user.is_admin or (str(user.user_id) in upload.viewers))):
        status_code = 403
        data = dict(
            code=status_code,
            name="Forbidden",
            description="User does not have access to this upload.")
        response = jsonify(data)
        response.status_code = status_code
        return response


@property  # type: ignore
def specs_url(self):
    '''
    Fixes issue where swagger-ui makes a call to swagger.json over HTTP.
    This can ONLY be used on servers that actually use HTTPS.  On servers that use HTTP,
    this code should not be used at all.
    '''
    return url_for(self.endpoint('specs'), _external=True, _scheme='https')


app = Flask(__name__)
''' The Flask app that serves all APIs. '''

CORS(app)

app.register_blueprint(docs_blueprint, url_prefix='/docs')
app.register_blueprint(dist_blueprint, url_prefix='/dist')
app.register_blueprint(gui_blueprint, url_prefix='/gui')

app.config["H5_BASE_DIR"] = config.fs.staging
app.before_request_funcs = {
    "h5grove": [auth_before_request]
}
app.register_blueprint(h5grove_blueprint, url_prefix='/h5grove')


@app.route('/alive')
def alive():
    ''' Simple endpoint to utilize kubernetes liveness/readiness probing. '''
    return "I am, alive!"



#!/usr/bin/env python3
"""Creates an instantiated NXDL schema XML tree by walking the dictionary nest

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import xml.etree.ElementTree as ET

from pyaml import yaml

from nexusparser.tools import nexus


NX_CLSS = nexus.get_nx_classes()
NX_NEW_DEFINED_CLASSES = ['NX_COMPLEX']
NX_TYPE_KEYS = nexus.get_nx_attribute_type()
NX_ATTR_IDNT = '\\@'
NX_UNIT_IDNT = 'unit'
NX_UNIT_TYPS = nexus.get_nx_units()


def yml_reader(inputfile):
    """
    Yaml module based reading of .yml file
    """
    with open(inputfile) as stream:
        return yaml.safe_load(stream)


def nx_name_type_resolving(tmp):
    """
    extracts the eventually custom name {optional_string}
    and type {nexus_type} from a YML section string.
    YML section string syntax: optional_string(nexus_type)
    """
    if tmp.count('(') == 1 and tmp.count(')') == 1:
        # we can safely assume that every valid YML key resolves
        # either an nx_ (type, base, candidate) class contains only 1 '(' and ')'
        index_start = tmp.index('(')
        index_end = tmp.index(')', index_start + 1)
        typ = tmp[index_start + 1:index_end]
        nam = tmp.replace('(' + typ + ')', '')
        return nam, typ
    # or a name for a member
    typ = ''
    nam = tmp
    return nam, typ


def xml_handle_doc(obj, value: str):
    """This function creates a 'doc' element instance, and appends it to an existing element

    """
    doctag = ET.SubElement(obj, 'doc')
    doctag.text = value


def xml_handle_units(obj, value):
    """This function creates a 'units' element instance, and appends it to an existing element

    """
    obj.set('units', value)


def xml_handle_exists(obj, value):
    """This function creates an 'exists' element instance, and appends it to an existing element

    """
    assert value is not None, 'xml_handle_exists, value must not be None!'
    if isinstance(value, list):
        if len(value) == 2 and value[0] == 'min':
            obj.set('minOccurs', str(value[1]))
        elif len(value) == 2 and value[0] == 'max':
            obj.set('maxOccurs', str(value[1]))
        elif len(value) == 4 and value[0] == 'min' and value[2] == 'max':
            obj.set('minOccurs', str(value[1]))
            if str(value[3]) != 'infty':
                obj.set('maxOccurs', str(value[3]))
            else:
                obj.set('maxOccurs', 'unbounded')
        elif len(value) == 4 and (value[0] != 'min' or value[2] != 'max'):
            raise ValueError('exists keyword needs to go either with an optional \
[recommended] list with two entries either [min, <uint>] or \
[max, <uint>], or a list of four entries [min, <uint>, max, <uint>] !')
        else:
            raise ValueError('exists keyword needs to go either with optional, \
recommended, a list with two entries either [min, <uint>] or \
[max, <uint>], or a list of four entries [min, <uint>, max, <uint>] !')
    else:
        if value == 'optional':
            obj.set('optional', 'true')
        elif value == 'recommended':
            obj.set('recommended', 'true')
        elif value == 'required':
            obj.set('minOccurs', '1')
        else:
            obj.set('minOccurs', '0')


def xml_handle_group(verbose, obj, value, keyword_name, keyword_type):
    """The function deals with group instances

"""
    grp = ET.SubElement(obj, 'group')
    if keyword_name != '':  # use the custom name for the group
        grp.set('name', keyword_name)
    grp.set('type', keyword_type)
    if isinstance(value, dict) and value != {}:
        recursive_build(grp, value, verbose)


def xml_handle_dimensions(obj, value: dict):
    """This function creates a 'dimensions' element instance, and appends it to an existing element

    """
    assert 'rank' in value.keys() and 'dim' in value.keys(), 'xml_handle_dimensions \
rank and/or dim not keys in value dict!'
    dims = ET.SubElement(obj, 'dimensions')
    dims.set('rank', str(value['rank']))
    for element in value['dim']:
        assert isinstance(element, list), 'xml_handle_dimensions, element is not a list!'
        assert len(element) >= 2, 'xml_handle_dimensions, list element has less than two entries!'
        dim = ET.SubElement(dims, 'dim')
        dim.set('index', str(element[0]))
        dim.set('value', str(element[1]))
        if len(element) == 3:
            assert element[2] == 'optional', 'xml_handle_dimensions element is \
a list with unexpected number of entries!'
            dim.set('required', 'false')


def xml_handle_enumeration(obj, value, verbose):
    """This function creates an 'enumeration' element instance.

Two cases are handled:
1) the items are in a list
2) the items are dictionaries and may contain a nested doc

"""
    enum = ET.SubElement(obj, 'enumeration')
    assert len(value) >= 1, 'xml_handle_enumeration, value must not be an empty list!'
    if isinstance(value, list):
        for element in value:
            itm = ET.SubElement(enum, 'item')
            itm.set('value', str(element))
    if isinstance(value, dict) and value != {}:
        for element in value.keys():
            itm = ET.SubElement(enum, 'item')
            itm.set('value', str(element))
            recursive_build(itm, value[str(element)], verbose)


def xml_handle_link(obj, keyword, value):
    """If we have an NXDL link we decode the name attribute from <optional string>(link)[:-6]

    """
    if len(keyword[:-6]) >= 1 and isinstance(value, dict) and 'target' in value.keys():
        if isinstance(value['target'], str) and len(value['target']) >= 1:
            lnk = ET.SubElement(obj, 'link')
            lnk.set('name', keyword[:-6])
            lnk.set('target', value['target'])
        else:
            raise ValueError(keyword + ' value for target member of a link is invalid !')
    else:
        raise ValueError(keyword + ' the formatting of what seems to be a link \
is invalid in the yml file !')


def xml_handle_symbols(obj, value: dict):
    """Handle a set of NXDL symbols as a child to obj

    """
    assert len(list(value.keys())) >= 1, 'xml_handle_symbols, symbols tables must not be empty!'
    syms = ET.SubElement(obj, 'symbols')
    if 'doc' in value.keys():
        doctag = ET.SubElement(syms, 'doc')
        doctag.text = value['doc']
    for kkeyword, vvalue in value.items():
        if kkeyword != 'doc':
            assert vvalue is not None and isinstance(vvalue, str), 'Put a comment in doc string!'
            sym = ET.SubElement(syms, 'symbol')
            sym.set('name', kkeyword)
            sym_doc = ET.SubElement(sym, 'doc')
            sym_doc.text = vvalue


def check_keyword_variable(verbose, keyword_name, keyword_type, value):
    """Check whether both keyword_name and keyword_type are empty, and complains if it is the case

"""
    if verbose:
        sys.stdout.write(f'{keyword_name}({keyword_type}): value type is {type(value)}\n')
    if keyword_name == '' and keyword_type == '':
        raise ValueError('Found an improper YML key !')


def helper_keyword_type(kkeyword_type):
    """This function is returning a value of keyword_type if it belong to NX_TYPE_KEYS

"""
    if kkeyword_type in NX_TYPE_KEYS:
        return kkeyword_type
    return None


def verbose_flag(verbose, keyword, value):
    """Verbose stdout printing for nested levels of yaml file, if verbose flag is active

"""
    if verbose:
        sys.stdout.write(f'  key:{keyword}; value type is {type(value)}\n')


def second_nested_level_handle(verbose, fld, value):
    """When a second dictionary is found inside a value, a new cycle of handlings is run

"""
    if isinstance(value, dict):
        for kkeyword, vvalue in iter(value.items()):
            verbose_flag(verbose, kkeyword, vvalue)
            if kkeyword[0:2] == NX_ATTR_IDNT:
                attr = ET.SubElement(fld, 'attribute')
                # attributes may also come with an nx_type specifier
                # which we need to decipher first
                kkeyword_name, kkeyword_type = \
                    nx_name_type_resolving(kkeyword[2:])
                attr.set('name', kkeyword_name)
                # typ = 'NX_CHAR'
                typ = helper_keyword_type(kkeyword_type) or 'NX_CHAR'
                attr.set('type', typ)
                if isinstance(vvalue, dict):
                    for kkkeyword, vvvalue in iter(vvalue.items()):
                        third_nested_level_handle(verbose, attr, kkeyword, kkkeyword, vvvalue)
            elif kkeyword == 'doc':
                xml_handle_doc(fld, vvalue)
            elif kkeyword == NX_UNIT_IDNT:
                xml_handle_units(fld, vvalue)
            elif kkeyword == 'exists':
                xml_handle_exists(fld, vvalue)
            elif kkeyword == 'dimensions':
                xml_handle_dimensions(fld, vvalue)
            elif kkeyword == 'enumeration':
                xml_handle_enumeration(fld, vvalue, verbose)
            elif kkeyword == 'link':
                fld.set('link', '')
            else:
                raise ValueError(kkeyword, ' faced unknown situation !')


def third_nested_level_handle(verbose, attr, kkeyword, kkkeyword, vvvalue):
    """When a third dictionary is found inside a value, a new cycle of handlings is run

"""
    verbose_flag(verbose, kkkeyword, vvvalue)
    if kkkeyword == 'doc':
        xml_handle_doc(attr, vvvalue)
    elif kkkeyword == 'exists':
        xml_handle_exists(attr, vvvalue)
    elif kkkeyword == 'enumeration':
        xml_handle_enumeration(attr, vvvalue, verbose)
    else:
        raise ValueError(
            kkeyword, kkkeyword, ' attribute handling !')


def attribute_attributes_handle(verbose, obj, value, keyword_name):
    """Handle the attributes found connected to attribute field"""
    # as an attribute identifier
    attr = ET.SubElement(obj, 'attribute')
    attr.set('name', keyword_name[2:])
    if value is not None:
        assert isinstance(value, dict), 'the keyword is an attribute, \
its value must be a dict!'
        for kkeyword, vvalue in iter(value.items()):
            verbose_flag(verbose, kkeyword, vvalue)
            if kkeyword == 'name':
                attr.set('name', vvalue)
            elif kkeyword == 'doc':
                xml_handle_doc(attr, vvalue)
            elif kkeyword == 'type':
                attr.set('type', vvalue.upper())
            elif kkeyword == 'enumeration':
                xml_handle_enumeration(attr, vvalue, verbose)
            elif kkeyword == 'exists':
                xml_handle_exists(attr, vvalue)
            else:
                raise ValueError(kkeyword + ' facing an unknown situation \
while processing attributes of an attribute ! Node tag:', obj.tag, 'Node content:', obj.attrib)
# handle special keywords (symbols),
# assumed that you do not encounter further symbols nested inside


def second_level_attributes_handle(fld, keyword, value):
    """If value is not a dictionary, this function handles the attributes of a nested field

"""
    if not isinstance(value, dict):
        if keyword == 'doc':
            xml_handle_doc(fld, value)
        elif keyword == NX_UNIT_IDNT:
            xml_handle_units(fld, value)
        elif keyword[0:2] == NX_ATTR_IDNT:  # attribute of a field
            raise ValueError(keyword, ' unknown attribute \
    of a field case coming from no dict !')
        elif keyword == 'exists':
            xml_handle_exists(fld, value)
        elif keyword == 'dimensions':
            raise ValueError(keyword, ' unknown dimensions \
    of a field case coming from no dict !')
        else:
            pass


def not_empty_keyword_name_handle(obj, keyword_type, keyword_name):
    """Handle a field in yaml file.
When a keyword is NOT:
symbol,
NX baseclass member,
attribute (\\@),
doc,
enumerations,
dimension,
exists,
then the not empty keyword_name is a field!
This simple function will define a new node of xml tree

"""
    typ = 'NX_CHAR'
    if keyword_type in NX_TYPE_KEYS + NX_NEW_DEFINED_CLASSES:
        typ = keyword_type
    # assume type is NX_CHAR, a NeXus default assumption if in doubt
    fld = ET.SubElement(obj, 'field')
    fld.set('name', keyword_name)
    fld.set('type', typ)
    return fld


def recursive_build(obj, dct, verbose):
    """obj is the current node of the XML tree where we want to append to,
    dct is a dictionary object which represents the content of a child to obj
    dct may contain further dictionary nests, representing NXDL groups,
    which trigger recursive processing
    NXDL fields may contain attributes but trigger no recursion so attributes are leafs.

    """
    for keyword, value in iter(dct.items()):
        keyword_name, keyword_type = nx_name_type_resolving(keyword)

        check_keyword_variable(verbose, keyword_name, keyword_type, value)
        if verbose:
            sys.stdout.write(f'keyword_name:{keyword_name} keyword_type {keyword_type}\n')
        if keyword[-6:] == '(link)':
            xml_handle_link(obj, keyword, value)

        elif keyword_type == '' and keyword_name == 'symbols':
            # print(value.key(), type(value.key()), value.value(), type(value.value()))
            xml_handle_symbols(obj, value)

        elif (keyword_type in NX_CLSS) or \
             (keyword_type not in NX_TYPE_KEYS + [''] + NX_NEW_DEFINED_CLASSES):
            # we can be sure we need to instantiate a new group
            xml_handle_group(verbose, obj, value, keyword_name, keyword_type)

        elif keyword_name[0:2] == NX_ATTR_IDNT:  # check if obj qualifies
            attribute_attributes_handle(verbose, obj, value, keyword_name)
        elif keyword == 'doc':
            xml_handle_doc(obj, value)

        elif keyword == 'enumeration':
            xml_handle_enumeration(obj, value, verbose)

        elif keyword == 'dimensions':
            xml_handle_dimensions(obj, value)

        elif keyword == 'exists':
            xml_handle_exists(obj, value)

        elif keyword_name != '':
            fld = not_empty_keyword_name_handle(obj, keyword_type, keyword_name)
            second_nested_level_handle(verbose, fld, value)
            second_level_attributes_handle(fld, keyword, value)
        else:
            pass



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import typing

from nomad import config

from .admin import admin


def _run_parallel(uploads, parallel: int, callable, label: str, print_progress: int = 0):
    import threading
    import time

    from nomad import utils, processing as proc

    if isinstance(uploads, (tuple, list)):
        uploads_count = len(uploads)

    else:
        uploads_count = uploads.count()
        uploads = list(uploads)  # copy the whole mongo query set to avoid cursor timeouts

    cv = threading.Condition()
    threads: typing.List[threading.Thread] = []

    state = dict(
        completed_count=0,
        skipped_count=0,
        available_threads_count=parallel)

    logger = utils.get_logger(__name__)

    print('%d uploads selected, %s ...' % (uploads_count, label))

    def process_upload(upload: proc.Upload):
        logger.info('%s started' % label, upload_id=upload.upload_id)

        completed = False
        try:
            if callable(upload, logger):
                completed = True
        except Exception as e:
            completed = True
            logger.error('%s failed' % label, upload_id=upload.upload_id, exc_info=e)

        with cv:
            state['completed_count'] += 1 if completed else 0
            state['skipped_count'] += 1 if not completed else 0
            state['available_threads_count'] += 1

            print(
                '   %s %s and skipped %s of %s uploads' %
                (label, state['completed_count'], state['skipped_count'], uploads_count))

            cv.notify()

    for upload in uploads:
        logger.info(
            'cli schedules parallel %s processing for upload' % label,
            current_process=upload.current_process,
            last_status_message=upload.last_status_message, upload_id=upload.upload_id)
        with cv:
            cv.wait_for(lambda: state['available_threads_count'] > 0)
            state['available_threads_count'] -= 1
            thread = threading.Thread(target=lambda: process_upload(upload))
            threads.append(thread)
            thread.start()

    def print_progress_lines():
        while True:
            time.sleep(print_progress)
            print('.', flush=True)

    if print_progress > 0:
        progress_thread = threading.Thread(target=print_progress_lines)
        progress_thread.daemon = True
        progress_thread.start()

    for thread in threads:
        thread.join()


def _run_processing(
        uploads, parallel: int, process, label: str, process_running: bool = False,
        wait_until_complete: bool = True, reset_first: bool = False, **kwargs):

    from nomad import processing as proc

    def run_process(upload, logger):
        logger.info(
            'cli calls %s processing' % label,
            current_process=upload.current_process,
            last_status_message=upload.last_status_message, upload_id=upload.upload_id)
        if upload.process_running and not process_running:
            logger.warn(
                'cannot trigger %s, since the upload is already/still processing' % label,
                current_process=upload.current_process,
                last_status_message=upload.last_status_message, upload_id=upload.upload_id)
            return False

        if reset_first:
            upload.reset(force=True)
        elif upload.process_running:
            upload.reset(force=True, process_status=proc.ProcessStatus.FAILURE)

        process(upload)
        if wait_until_complete:
            upload.block_until_complete(interval=.5)
        else:
            upload.block_until_complete_or_waiting_for_result(interval=.5)

        if upload.process_status == proc.ProcessStatus.FAILURE:
            logger.info('%s with failure' % label, upload_id=upload.upload_id)

        logger.info('%s complete' % label, upload_id=upload.upload_id)
        return True

    _run_parallel(uploads, parallel=parallel, callable=run_process, label=label, **kwargs)


@admin.group(help='Upload related commands')
@click.option('--uploads-mongo-query', type=str, help='A query')
@click.option('--entries-mongo-query', type=str, help='A query')
@click.option('--entries-es-query', type=str, help='A query')
@click.option('--unpublished', help='Select only uploads in staging', is_flag=True)
@click.option('--published', help='Select only uploads that are publised', is_flag=True)
@click.option('--outdated', help='Select published uploads with older nomad version', is_flag=True)
@click.option('--processing', help='Select only processing uploads', is_flag=True)
@click.option('--processing-failure-uploads', is_flag=True, help='Select uploads with failed processing')
@click.option('--processing-failure-entries', is_flag=True, help='Select uploads with entries with failed processing')
@click.option('--processing-failure', is_flag=True, help='Select uploads where the upload or any entry has failed processing')
@click.option('--processing-incomplete-uploads', is_flag=True, help='Select uploads that have not yet been processed')
@click.option('--processing-incomplete-entries', is_flag=True, help='Select uploads where any entry has net yot been processed')
@click.option('--processing-incomplete', is_flag=True, help='Select uploads where the upload or any entry has not yet been processed')
@click.option('--processing-necessary', is_flag=True, help='Select uploads where the upload or any entry has either not been processed or processing has failed in the past')
@click.option('--unindexed', is_flag=True, help='Select uploads that have no entries in the elastic search index.')
@click.pass_context
def uploads(ctx, **kwargs):
    ctx.obj.uploads_kwargs = kwargs


def _query_uploads(
        uploads,
        unpublished: bool, published: bool, processing: bool, outdated: bool,
        uploads_mongo_query: str, entries_mongo_query: str, entries_es_query: str,
        processing_failure_uploads: bool, processing_failure_entries: bool,
        processing_failure: bool, processing_incomplete_uploads: bool,
        processing_incomplete_entries: bool, processing_incomplete: bool,
        processing_necessary: bool, unindexed: bool):

    '''
    Produces a list of uploads (mongoengine proc.Upload objects) based on a given
    list of upoad ids and further filter parameters.
    '''

    from typing import Set, cast
    import json
    from mongoengine import Q

    from nomad import infrastructure, processing as proc, search
    from nomad.app.v1 import models

    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    if uploads is not None and len(uploads) == 0:
        uploads = None  # None meaning all uploads
    else:
        uploads = set(uploads)

    entries_mongo_query_q = Q()
    if entries_mongo_query:
        entries_mongo_query_q = Q(**json.loads(entries_mongo_query))

    entries_query_uploads: Set[str] = None

    if entries_es_query is not None:
        entries_es_query_dict = json.loads(entries_es_query)
        results = search.search(
            owner='admin',
            query=entries_es_query_dict,
            pagination=models.MetadataPagination(page_size=0),
            user_id=config.services.admin_user_id,
            aggregations={
                'uploads': models.Aggregation(
                    terms=models.TermsAggregation(
                        quantity='upload_id',
                        pagination=models.AggregationPagination(
                            page_size=10000
                        )
                    )
                )
            })

        entries_query_uploads = set([
            cast(str, bucket.value)
            for bucket in results.aggregations['uploads'].terms.data])  # pylint: disable=no-member

    if outdated:
        entries_mongo_query_q &= Q(nomad_version={'$ne': config.meta.version})

    if processing_failure_entries or processing_failure or processing_necessary:
        entries_mongo_query_q &= Q(process_status=proc.ProcessStatus.FAILURE)

    if processing_incomplete_entries or processing_incomplete or processing_necessary:
        entries_mongo_query_q &= Q(process_status__in=proc.ProcessStatus.STATUSES_PROCESSING)

    mongo_entry_based_uploads = set(proc.Entry.objects(entries_mongo_query_q).distinct(field="upload_id"))
    if entries_query_uploads is not None:
        entries_query_uploads = entries_query_uploads.intersection(mongo_entry_based_uploads)
    else:
        entries_query_uploads = mongo_entry_based_uploads

    if entries_query_uploads:
        uploads_mongo_query_q = Q(upload_id__in=list(entries_query_uploads))
    else:
        uploads_mongo_query_q = Q()

    if uploads_mongo_query:
        uploads_mongo_query_q &= Q(**json.loads(uploads_mongo_query))

    if published:
        uploads_mongo_query_q &= Q(publish_time__exists=True)

    if unpublished:
        uploads_mongo_query_q &= Q(publish_time__exists=False)

    if processing:
        uploads_mongo_query_q &= Q(process_status__in=proc.ProcessStatus.STATUSES_PROCESSING)

    if processing_failure_uploads or processing_failure or processing_necessary:
        uploads_mongo_query_q &= Q(process_status=proc.ProcessStatus.FAILURE)

    if processing_incomplete_uploads or processing_incomplete or processing_necessary:
        uploads_mongo_query_q &= Q(process_status__in=proc.ProcessStatus.STATUSES_PROCESSING)

    final_query = uploads_mongo_query_q
    if uploads is not None:
        final_query &= Q(upload_id__in=list(uploads))

    return final_query, proc.Upload.objects(final_query)


@uploads.command(help='List selected uploads')
@click.argument('UPLOADS', nargs=-1)
@click.option('-e', '--entries', is_flag=True, help='Show details about entries.')
@click.option('--ids', is_flag=True, help='Only show a list of ids.')
@click.option('--json', is_flag=True, help='Output a JSON array of ids.')
@click.pass_context
def ls(ctx, uploads, entries, ids, json):
    import tabulate

    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    def row(upload):
        row = [
            upload.upload_id,
            upload.upload_name,
            upload.main_author,
            upload.process_status,
            upload.published]

        if entries:
            row += [
                upload.total_entries_count,
                upload.failed_entries_count,
                upload.total_entries_count - upload.processed_entries_count]

        return row

    headers = ['id', 'upload_name', 'user', 'process', 'published']
    if entries:
        headers += ['entries', 'failed', 'processing']

    if ids:
        for upload in uploads:
            print(upload.upload_id)
        return

    if json:
        print('[%s]' % ','.join(['"%s"' % upload.upload_id for upload in uploads]))
        return

    print('%d uploads selected, showing no more than first 10' % uploads.count())
    print(tabulate.tabulate(
        [row(upload) for upload in uploads[:10]],
        headers=headers))


@uploads.command(help='Change the owner of the upload and all its entries.')
@click.argument('USERNAME', nargs=1)
@click.argument('UPLOADS', nargs=-1)
@click.pass_context
def chown(ctx, username, uploads):
    from nomad import datamodel

    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    print('%d uploads selected, changing owner ...' % uploads.count())

    user = datamodel.User.get(username=username)
    for upload in uploads:
        upload.edit_upload_metadata(
            edit_request_json=dict(metadata={'main_author': user.user_id}),
            user_id=config.services.admin_user_id)


@uploads.command(help='Reset the processing state.')
@click.argument('UPLOADS', nargs=-1)
@click.option('--with-entries', is_flag=True, help='Also reset all entries.')
@click.option('--success', is_flag=True, help='Set the process status to success instead of pending')
@click.option('--failure', is_flag=True, help='Set the process status to failure instead of pending.')
@click.pass_context
def reset(ctx, uploads, with_entries, success, failure):
    from nomad import processing as proc

    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)
    uploads_count = uploads.count()

    print('%d uploads selected, resetting their processing ...' % uploads_count)

    i = 0
    for upload in uploads:
        if with_entries:
            entry_update = proc.Entry.reset_pymongo_update()
            if success:
                entry_update['process_status'] = proc.ProcessStatus.SUCCESS
            if failure:
                entry_update['process_status'] = proc.ProcessStatus.FAILURE

            proc.Entry._get_collection().update_many(
                dict(upload_id=upload.upload_id), {'$set': entry_update})

        upload.reset(force=True)
        if success:
            upload.process_status = proc.ProcessStatus.SUCCESS
        if failure:
            upload.process_status = proc.ProcessStatus.FAILURE
        upload.save()
        i += 1
        print('resetted %d of %d uploads' % (i, uploads_count))


@uploads.command(help='(Re-)index all entries of the given uploads.')
@click.argument('UPLOADS', nargs=-1)
@click.option('--parallel', default=1, type=int, help='Use the given amount of parallel processes. Default is 1.')
@click.option('--transformer', help='Qualified name to a Python function that should be applied to each EntryMetadata.')
@click.option('--skip-materials', is_flag=True, help='Only update the entries index.')
@click.option('--print-progress', default=0, type=int, help='Prints a dot every given seconds. Can be used to keep terminal open that have an i/o-based timeout.')
@click.pass_context
def index(ctx, uploads, parallel, transformer, skip_materials, print_progress):
    from nomad import search

    transformer_func = None
    if transformer is not None:
        import importlib
        module_name, func_name = transformer.rsplit('.', 1)
        module = importlib.import_module(module_name)
        transformer_func = getattr(module, func_name)

    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    def transform(entries):
        for entry in entries:
            try:
                entry = transformer_func(entry)
            except Exception as e:
                import traceback
                traceback.print_exc()
                print(f'   ERROR failed to transform entry (stop transforming for upload): {str(e)}')
                break

    def index_upload(upload, logger):
        with upload.entries_metadata() as entries:
            if transformer is not None:
                transform(entries)
            archives = [entry.m_parent for entry in entries]
            search.index(archives, update_materials=not skip_materials, refresh=True)

        return True

    _run_parallel(uploads, parallel, index_upload, 'index', print_progress=print_progress)


def delete_upload(upload, skip_es: bool = False, skip_files: bool = False, skip_mongo: bool = False):
    from nomad import search, files, utils, processing as proc

    # delete elastic
    if not skip_es:
        search.delete_upload(upload_id=upload.upload_id, update_materials=True, refresh=True)

    # delete files
    if not skip_files:
        # do it twice to get the two potential versions 'public' and 'staging'
        for _ in range(0, 2):
            upload_files = files.UploadFiles.get(upload_id=upload.upload_id)

            try:
                if upload_files is not None:
                    upload_files.delete()
            except Exception as e:
                logger = utils.get_logger(__name__)
                logger.error('could not delete files', exc_info=e)
                break

    # delete mongo
    if not skip_mongo:
        proc.Entry.objects(upload_id=upload.upload_id).delete()
        upload.delete()


@uploads.command(help='Delete selected upload')
@click.argument('UPLOADS', nargs=-1)
@click.option('--skip-es', help='Keep the elastic index version of the data.', is_flag=True)
@click.option('--skip-mongo', help='Keep uploads and entries in mongo.', is_flag=True)
@click.option('--skip-files', help='Keep all related files.', is_flag=True)
@click.pass_context
def rm(ctx, uploads, skip_es, skip_mongo, skip_files):
    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    print('%d uploads selected, deleting ...' % uploads.count())

    for upload in uploads:
        delete_upload(upload, skip_es=skip_es, skip_mongo=skip_mongo, skip_files=skip_files)


@uploads.command(help='Reprocess selected uploads.')
@click.argument('UPLOADS', nargs=-1)
@click.option('--parallel', default=1, type=int, help='Use the given amount of parallel processes. Default is 1.')
@click.option('--process-running', is_flag=True, help='Also reprocess already running processes.')
@click.option('--setting', type=str, multiple=True, help='key=value to overwrite a default reprocess config setting.')
@click.option('--print-progress', default=0, type=int, help='Prints a dot every given seconds. Can be used to keep terminal open that have an i/o-based timeout.')
@click.pass_context
def process(ctx, uploads, parallel: int, process_running: bool, setting: typing.List[str], print_progress: int):
    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)
    settings: typing.Dict[str, bool] = {}
    for settings_str in setting:
        key, value = settings_str.split('=')
        settings[key] = bool(value)
    _run_processing(
        uploads, parallel, lambda upload: upload.process_upload(reprocess_settings=settings),
        'processing', process_running=process_running, reset_first=True, print_progress=print_progress)


@uploads.command(help='Repack selected uploads.')
@click.argument('UPLOADS', nargs=-1)
@click.pass_context
def re_pack(ctx, uploads):
    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    for upload in uploads:
        if not upload.published:
            print(f'Cannot repack unpublished upload {upload.upload_id}')
            continue

        upload.upload_files.re_pack(upload.with_embargo)
        print(f'successfully re-packed {upload.upload_id}')


@uploads.command(help='Attempt to abort the processing of uploads.')
@click.argument('UPLOADS', nargs=-1)
@click.option('--entries', is_flag=True, help='Only stop entries processing.')
@click.option('--kill', is_flag=True, help='Use the kill signal and force task failure.')
@click.option('--no-celery', is_flag=True, help='Do not attempt to stop the actual celery tasks')
@click.pass_context
def stop(ctx, uploads, entries: bool, kill: bool, no_celery: bool):
    import mongoengine

    from nomad import utils, processing as proc
    query, _ = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    logger = utils.get_logger(__name__)

    def stop_all(query):
        for process in query:
            logger_kwargs = dict(upload_id=process.upload_id)
            if isinstance(process, proc.Entry):
                logger_kwargs.update(entry_id=process.entry_id)

            if not no_celery:
                logger.info(
                    'send terminate celery task', celery_task_id=process.celery_task_id,
                    kill=kill, **logger_kwargs)

            kwargs = {}
            if kill:
                kwargs.update(signal='SIGKILL')
            try:
                if not no_celery:
                    proc.app.control.revoke(process.celery_task_id, terminate=True, **kwargs)
            except Exception as e:
                logger.warning(
                    'could not revoke celery task', exc_info=e,
                    celery_task_id=process.celery_task_id, **logger_kwargs)

            if kill:
                logger.info(
                    'fail proc', celery_task_id=process.celery_task_id, kill=kill,
                    **logger_kwargs)

                process.fail('process terminate via nomad cli')

    running_query = query & mongoengine.Q(process_status__in=proc.ProcessStatus.STATUSES_PROCESSING)
    stop_all(proc.Entry.objects(running_query))
    if not entries:
        stop_all(proc.Upload.objects(running_query))


@uploads.group(help='Check certain integrity criteria')
@click.pass_context
def integrity(ctx):
    pass


@integrity.command(help='Uploads that have more entries in mongo than in ES.')
@click.argument('UPLOADS', nargs=-1)
@click.pass_context
def entry_index(ctx, uploads):
    from nomad.search import search
    from nomad.processing import Upload
    from nomad.app.v1.models import Pagination

    _, uploads = _query_uploads(uploads, **ctx.obj.uploads_kwargs)

    upload: Upload = None
    for upload in uploads:
        search_results = search(
            owner='admin',
            query=dict(upload_id=upload.upload_id),
            pagination=Pagination(page_size=0),
            user_id=config.services.admin_user_id)

        if search_results.pagination.total != upload.total_entries_count:
            print(upload.upload_id)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_elastic_section_strain_diagrams(MSection):
    '''
    section collecting the data of the strain diagrams
    '''

    m_def = Section(validate=False)

    x_elastic_strain_diagram_values = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations', 'x_elastic_strain_diagram_number_of_eta'],
        description='''
        Values of the energy(units:J)/d2E(units:Pa)/cross-validation (depending on the
        value of x_elastic_strain_diagram_type)
        ''')

    x_elastic_strain_diagram_eta_values = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations', 'x_elastic_strain_diagram_number_of_eta'],
        description='''
        eta values used the strain diagrams
        ''')

    x_elastic_strain_diagram_number_of_eta = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of strain values used in the strain diagram
        ''')

    x_elastic_strain_diagram_type = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of strain diagram. Possible values are: energy; cross-validation (cross-
        validation error); d2E (second derivative of the energy wrt the strain)
        ''')

    x_elastic_strain_diagram_polynomial_fit_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Order of the polynomial fit
        ''')


class x_elastic_section_fitting_parameters(MSection):
    '''
    section collecting the fitting parameters used to calculate the elastic constants
    '''

    m_def = Section(validate=False)

    x_elastic_fitting_parameters_eta = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations'],
        description='''
        eta values used to calculate the elastic constants
        ''')

    x_elastic_fitting_parameters_polynomial_order = Quantity(
        type=np.dtype(np.int32),
        shape=['x_elastic_number_of_deformations'],
        description='''
        polynomial order used to fit the Energy vs. volume curve and to calculate the
        elastic constants
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

#     x_elastic_elastic_constant_order = Quantity(
#         type=np.dtype(np.int32),
#         shape=[],
#         description='''
#         Order of the elastic constant
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_elastic_constant_order'))

#     x_elastic_number_of_deformations = Quantity(
#         type=np.dtype(np.int32),
#         shape=[],
#         description='''
#         number of deformed structures equally spaced in strain, which are generated
#         between the maximum negative strain and the maximum positive one
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_number_of_deformations'))

#     x_elastic_deformation_types = Quantity(
#         type=np.dtype('U'),
#         shape=['x_elastic_number_of_deformations', 6],
#         description='''
#         deformation types
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_deformation_types'))

#     x_elastic_calculation_method = Quantity(
#         type=str,
#         shape=[],
#         description='''
#         Method of calculation
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_calculation_method'))

#     x_elastic_code = Quantity(
#         type=str,
#         shape=[],
#         description='''
#         Code used for the calculation of the elastic constants
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_code'))

#     x_elastic_max_lagrangian_strain = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         description='''
#         Maximum lagrangian strain used to calculate the elastic constants
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_max_lagrangian_strain'))

#     x_elastic_number_of_distorted_structures = Quantity(
#         type=np.dtype(np.int32),
#         shape=[],
#         description='''
#         Number of distorted structures used to calculate the elastic constants
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_number_of_distorted_structures'))

    x_elastic_section_fitting_parameters = SubSection(
        sub_section=SectionProxy('x_elastic_section_fitting_parameters'),
        repeats=True)


# class section_single_configuration_calculation(public.section_single_configuration_calculation):

#     m_def = Section(validate=False, extends_base_section=True)

#     x_elastic_2nd_order_constants_notation_matrix = Quantity(
#         type=np.dtype('U'),
#         shape=[6, 6],
#         description='''
#         Symmetry of the second-order elastic constant matrix in Voigt notation
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_notation_matrix'))

#     x_elastic_2nd_order_constants_matrix = Quantity(
#         type=np.dtype(np.float64),
#         shape=[6, 6],
#         unit='pascal',
#         description='''
#         2nd order elastic constant (stiffness) matrix in GPa
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_matrix'))

#     x_elastic_3rd_order_constants_matrix = Quantity(
#         type=np.dtype(np.float64),
#         shape=[6, 6, 6],
#         unit='pascal',
#         description='''
#         3rd order elastic constant (stiffness) matrix in GPa
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_3rd_order_constants_matrix'))

#     x_elastic_2nd_order_constants_compliance_matrix = Quantity(
#         type=np.dtype(np.float64),
#         shape=[6, 6],
#         unit='1 / pascal',
#         description='''
#         Elastic compliance matrix in 1/GPa
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_compliance_matrix'))

#     x_elastic_Voigt_bulk_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Voigt bulk modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Voigt_bulk_modulus'))

#     x_elastic_Voigt_shear_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Voigt shear modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Voigt_shear_modulus'))

#     x_elastic_Reuss_bulk_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Reuss bulk modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Reuss_bulk_modulus'))

#     x_elastic_Reuss_shear_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Reuss shear modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Reuss_shear_modulus'))

#     x_elastic_Hill_bulk_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Hill bulk modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Hill_bulk_modulus'))

#     x_elastic_Hill_shear_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Hill shear modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Hill_shear_modulus'))

#     x_elastic_Voigt_Young_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Voigt Young modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Voigt_Young_modulus'))

#     x_elastic_Voigt_Poisson_ratio = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         description='''
#         Voigt Poisson ratio
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Voigt_Poisson_ratio'))

#     x_elastic_Reuss_Young_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Reuss Young modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Reuss_Young_modulus'))

#     x_elastic_Reuss_Poisson_ratio = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         description='''
#         Reuss Poisson ratio
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Reuss_Poisson_ratio'))

#     x_elastic_Hill_Young_modulus = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         unit='pascal',
#         description='''
#         Hill Young modulus
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Hill_Young_modulus'))

#     x_elastic_Hill_Poisson_ratio = Quantity(
#         type=np.dtype(np.float64),
#         shape=[],
#         description='''
#         Hill Poisson ratio
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_Hill_Poisson_ratio'))

#     x_elastic_eigenvalues = Quantity(
#         type=np.dtype(np.float64),
#         shape=[6],
#         unit='pascal',
#         description='''
#         Eigemvalues of the stiffness matrix
#         ''',
#         a_legacy=LegacyDefinition(name='x_elastic_eigenvalues'))

#     x_elastic_section_strain_diagrams = SubSection(
#         sub_section=SectionProxy('x_elastic_section_strain_diagrams'),
#         repeats=True,
#         a_legacy=LegacyDefinition(name='x_elastic_section_strain_diagrams'))


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_elastic_space_group_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Space-group number of the system
        ''')

    x_elastic_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='m ** 3',
        description='''
        Volume of the equilibrium unit cell
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Union, Optional
from fastapi import Depends, APIRouter, status, HTTPException, Query
from pydantic.main import BaseModel

from nomad import infrastructure, config, datamodel
from nomad.utils import strip

from .auth import create_user_dependency
from ..models import User, HTTPExceptionModel
from ..utils import create_responses

router = APIRouter()
default_tag = 'users'


_authentication_required_response = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. The operation requires authorization,
        but no or bad authentication credentials are given.''')}

_bad_invite_response = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The invite is invalid.''')}


class Users(BaseModel):
    data: List[User]


@router.get(
    '/me',
    tags=[default_tag],
    summary='Get your account data',
    description='Returns the account data of the authenticated user.',
    responses=create_responses(_authentication_required_response),
    response_model=User)
async def read_users_me(current_user: User = Depends(create_user_dependency(required=True))):
    return current_user


@router.get(
    '',
    tags=[default_tag],
    summary='Get existing users',
    description='Get existing users for given criteria',
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    response_model=Users)
async def get_users(
        prefix: Optional[str] = Query(None, description=strip('''
                Search the user with the given prefix.
            ''')),
        user_id: Union[List[str], None] = Query(None, description=strip('''
                To get the user(s) by their user_id(s).
            '''))):

    users: List[User] = []
    if user_id:
        if type(user_id) == list:
            for _user_id in user_id:
                try:
                    user = datamodel.User.get(user_id=str(_user_id)).m_copy()
                    user.email = None
                    users = users + [user]
                except KeyError:
                    pass
        else:
            user = datamodel.User.get(user_id=str(user_id)).m_copy()
            user.email = None
            users = [user]
    elif prefix:
        for user in infrastructure.keycloak.search_user(prefix):
            user_dict = user.m_to_dict(include_derived=True)
            user_dict['email'] = None
            users.append(user_dict)

    return dict(data=users)


@router.get(
    '/{user_id}',
    tags=[default_tag],
    summary='Get existing users',
    description='Get the user using the given user_id',
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    response_model=User)
async def get_user(user_id: str):
    user = datamodel.User.get(user_id=str(user_id)).m_copy()
    user.email = None
    return user


@router.put(
    '/invite',
    tags=[default_tag],
    summary='Invite a new user',
    responses=create_responses(_authentication_required_response, _bad_invite_response),
    response_model=User)
async def invite_user(user: User, current_user: User = Depends(create_user_dependency(required=True))):
    if config.keycloak.oasis:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='User invite does not work this NOMAD OASIS.')

    json_data = user.dict()
    try:
        user = datamodel.User.m_from_dict(json_data)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid user data: %s' % str(e))

    if user.email is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid user data: email is required')

    try:
        error = infrastructure.keycloak.add_user(user, invite=True)
    except KeyError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid user data: %s' % str(e))

    if error is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Could not invite user: %s' % str(error))

    return datamodel.User.get(username=user.username), 200



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict
from elasticsearch_dsl import Q
from cachetools import cached

from optimade.filterparser import LarkParser
from optimade.filtertransformers.elasticsearch import (
    Quantity, ElasticTransformer as OPTElasticTransformer, _cmp_operators)

from .common import provider_specific_fields


_parser = LarkParser(version=(0, 10, 1))


class FilterException(Exception):
    ''' Raised on parsing a filter expression with syntactic of semantic errors. '''
    pass


@cached(cache={})
def _get_transformer(without_prefix):
    from nomad.datamodel import OptimadeEntry
    quantities: Dict[str, Quantity] = {
        q.name: Quantity(
            q.name, es_field='optimade.%s' % q.name,
            elastic_mapping_type=q.a_elasticsearch.mapping['type'])

        for q in OptimadeEntry.m_def.all_quantities.values()
        if 'elasticsearch' in q.m_annotations}

    quantities['id'] = Quantity('id', es_field='entry_id', elastic_mapping_type='keyword')
    quantities['immutable_id'] = Quantity('immutable_id', es_field='entry_id', elastic_mapping_type='keyword')
    quantities['last_modified'] = Quantity(
        'last_modified', es_field='upload_create_time', elastic_mapping_type='date')

    quantities['elements'].length_quantity = quantities['nelements']
    quantities['elements'].nested_quantity = quantities['elements_ratios']
    quantities['elements_ratios'].nested_quantity = quantities['elements_ratios']

    for name, search_quantity in provider_specific_fields().items():
        names = ['_nmd_' + name]
        if without_prefix:
            names.append(name)

        for name in names:
            if name not in quantities:
                quantities[name] = Quantity(
                    name,
                    es_field=search_quantity.search_field,
                    elastic_mapping_type=search_quantity.mapping['type'])

    return ElasticTransformer(quantities=quantities.values())


def parse_filter(filter_str: str, without_prefix=False) -> Q:
    ''' Parses the given optimade filter str and returns a suitable elastic search query.

    Arguments:
        filter_str: Can be direct user input with no prior processing.
        nomad_properties: Also include the nomad proprietary properties.
        without_prefix: Do not prefix the nomad proprietary properties with _nmd_.

    Raises:
        FilterException: If the given str cannot be parsed, or if there are any semantic
            errors in the given expression.
    '''
    transformer = _get_transformer(without_prefix)

    try:
        parse_tree = _parser.parse(filter_str)
    except Exception as e:
        raise FilterException('Syntax error: %s' % str(e))

    try:
        query = transformer.transform(parse_tree)
    except Exception as e:
        raise FilterException('Semantic error: %s' % str(e))

    return query


class ElasticTransformer(OPTElasticTransformer):
    def _query_op(self, quantity, op, value, nested=None):
        """
        Return a range, match, or term query for the given quantity, comparison
        operator, and value
        """
        field = self._field(quantity, nested=nested)
        if op in _cmp_operators:
            return Q("range", **{field: {_cmp_operators[op]: value}})

        if quantity.elastic_mapping_type == 'text':
            query_type = "match"
        elif quantity.elastic_mapping_type in ['keyword', 'integer', 'float', 'bool']:
            query_type = "term"
        else:
            raise NotImplementedError("Quantity has unsupported ES field type")

        if op in ["=", ""]:
            return Q(query_type, **{field: value})

        if op == "!=":
            return ~Q(  # pylint: disable=invalid-unary-operand-type
                query_type, **{field: value}
            )

    def _has_query_op(self, quantities, op, predicate_zip_list):
        # We override this to add 'HAS ONLY' support.
        if op == 'HAS ONLY':
            # HAS ONLY can be achieved by rewriting to a combination of HAS ALL and
            # length = n_values. Therefore, it is only support for quantities with a
            # length quantity.
            if len(quantities) > 1:
                raise Exception('HAS ONLY is not supported with zip')
            quantity = quantities[0]

            if quantity.length_quantity is None:
                raise Exception('HAS ONLY is not supported by %s' % quantity.name)

            has_all = super()._has_query_op(quantities, 'HAS ALL', predicate_zip_list)
            has_length = Q('term', **{quantity.length_quantity.es_field: len(predicate_zip_list)})
            return has_all & has_length

        else:
            return super()._has_query_op(quantities, op, predicate_zip_list)

    def property_zip_addon(self, args):
        return args

    def value_zip(self, args):
        return self.value_list(args)

    def value_zip_list(self, args):
        return args
